package mbOrbits;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import javax.imageio.ImageIO;
import java.io.*;
import java.util.Vector;

public class MandelbrotCanvas extends JPanel implements ActionListener, ChangeListener {
	
	private BufferedImage OSI;
	private int imageWidth, imageHeight;
	

	private MandComp comp;
	private int[] palette = new int[260];  // only 253 used; extra space for rotating palette  -- now, only 252 are used!
	
	private double xmin, xmax, ymin, ymax;

	
	private int maxIterations = 250;
	
    private int pointsOnOrbit = 120;
    private int orbitLines = 50;
	private boolean dragging;
	private int startX, startY, endX, endY;
	private double X,Y;  // first point on orbit.
	private final static int NO_ORBIT = 0, POINT_ORBIT = 1, LINE_ORBIT = 2, MULTI_ORBIT = 3;
	private int orbitType = NO_ORBIT;
	private Vector orbitData = new Vector();
	private Color[] orbitPalette;
	
	private JTextField xInput;
	private JTextField yInput;
	private JSlider pointCountSlider;
	private JButton setButton;
	
	public MandelbrotCanvas() {
		this(800,600);
	}
	
	public MandelbrotCanvas(int width, int height) {
		imageWidth = width;
		imageHeight = height;
		setBackground(Color.lightGray);  
		setPreferredSize(new Dimension(imageWidth,imageHeight));
		setOpaque(true);
		for (int i = 0; i < 250; i++)
			palette[i] = new Color(125+i/2,125+i/2,125+i/2).getRGB();
		palette[250] = 0;  // Mandelbrot color
		palette[251] = 0xFFFFFF;  // formerly drag rect drawing color  -- no longer referenced anywhere, never changes
		palette[252] = Color.lightGray.getRGB();   // background color -- never changes
		orbitPalette = new Color[250];
		for (int i = 0; i < 250; i++)
			orbitPalette[i] = Color.getHSBColor((0.9F*i)/250,1,1);
		xmin = -2.333;
		xmax = 1.2;
		ymin = -1.25;
		ymax = 1.25;
		checkAspect();
		comp = new ColumnsMandComp();
		xInput = new JTextField(10);
		yInput = new JTextField(10);
		pointCountSlider = new JSlider(1,1000,120);
		setButton = new JButton("Set");
		xInput.addActionListener(this);
		yInput.addActionListener(this);
		pointCountSlider.addChangeListener(this);
		setButton.addActionListener(this);
		addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent evt) {
				dragging = false;
				startX = endX = evt.getX();
				startY = endY = evt.getY();
				X = xmin + startX*(xmax - xmin)/imageWidth;
				Y = ymax - startY*(ymax - ymin)/imageHeight;
				if (evt.isControlDown() || evt.isMetaDown()) {
					orbitType = LINE_ORBIT;
					setXYInputs(startX,startY);
				}
				else if (evt.isAltDown() || evt.isShiftDown()) {
					orbitType = MULTI_ORBIT;
					dragging = true;
				}
				else {
					orbitType = POINT_ORBIT;
					setXYInputs(startX,startY);
				}
				orbitData.setSize(0);
				repaint();
			}
			public void mouseReleased(MouseEvent evt) {
				if (! dragging)
					return;
				dragging = false;
				endX = evt.getX();
				endY = evt.getY();
				if (Math.abs(startX-endX) < 2 && (Math.abs(startY-endY) < 2))
					orbitType = NO_ORBIT;
				repaint();
			}
		});
		addMouseMotionListener(new MouseMotionAdapter() {
			public void mouseDragged(MouseEvent evt) {
				if (dragging) {
					endX = evt.getX();
					endY = evt.getY();
					if (Math.abs(startX-endX) < 2 && (Math.abs(startY-endY) < 2))
						orbitType = NO_ORBIT;
					repaint();
				}
				else if (orbitType == POINT_ORBIT || orbitType == LINE_ORBIT) {
					startX = endX = evt.getX();
					startY = endX = evt.getY();
					X = xmin + startX*(xmax - xmin)/imageWidth;
					Y = ymax - startY*(ymax - ymin)/imageHeight;
					orbitData.setSize(0);
					setXYInputs(startX,startY);
					repaint();
				}
			}
		});
	}
	
	
	public void actionPerformed(ActionEvent evt) {
		double x,y;
		try {
			x = Double.parseDouble(xInput.getText());
		}
		catch (NumberFormatException e) {
			JOptionPane.showMessageDialog(this, "The value specifed for x is not a legal real number.");
			return;
		}
		try {
			y = Double.parseDouble(yInput.getText());
		}
		catch (NumberFormatException e) {
			JOptionPane.showMessageDialog(this, "The value specifed for y is not a legal real number.");
			return;
		}
		if (x < -5 || x > 5) {
			JOptionPane.showMessageDialog(this, "The value specified for x must be in the range -5 to 5.");
			return;
		}
		if (y < -5 || y > 5) {
			JOptionPane.showMessageDialog(this, "The value specified for y must be in the range -5 to 5.");
			return;
		}
		if (orbitType == NO_ORBIT || orbitType == MULTI_ORBIT)
			orbitType = POINT_ORBIT;
		orbitData.setSize(0);
		X = x;
		Y = y;
		repaint();
	}
	
	public void stateChanged(ChangeEvent evt) {
		if (orbitType != NO_ORBIT) {
			pointsOnOrbit = pointCountSlider.getValue();
			orbitData.setSize(0);
			repaint();
		}
	}
	
	
	private void setXYInputs(int x, int y) {
		double X = xmin + x*(xmax-xmin)/imageWidth;
		double Y = ymax - y*(ymax-ymin)/imageHeight;
		xInput.setText("" + X);
		yInput.setText("" + Y);
	}
	
	public JPanel getControlPanel() {
		JPanel p = new JPanel();
		p.setLayout(new GridLayout(1,3,5,5));
		JPanel p1 = new JPanel();
		p1.setLayout(new BorderLayout(5,5));
		JPanel p2 = new JPanel();
		p2.setLayout(new BorderLayout(5,5));
		p.add(p1);
		p.add(p2);
		p.add(pointCountSlider);
		p1.add(xInput, BorderLayout.CENTER);
		p1.add(new JLabel(" x = "), BorderLayout.WEST);
		p2.add(yInput, BorderLayout.CENTER);
		p2.add(new JLabel(" y = "), BorderLayout.WEST);
		JPanel panel = new JPanel();
		panel.setLayout(new BorderLayout(5,5));
		panel.add(p,BorderLayout.CENTER);
		panel.add(setButton,BorderLayout.WEST);
		return panel;
	}
		
	private void forceStopDraw() {
		if (comp.isRunning()) {
			comp.cancel();
			try {
				for (int i = 0; i < 20 && comp.isRunning(); i++)
				   Thread.sleep(100);
			}
			catch (InterruptedException e) {
			}
		}
	}
	
	public void setCanvasSize(int width, int height) {
		if (width == imageWidth && height == imageHeight)
			return;
		forceStopDraw();
		OSI = null;
		imageHeight = height;
		imageWidth = width;
		checkAspect();
		setPreferredSize(new Dimension(width,height));
		makeImage();
		doDraw();
	}
	
	private void checkAspect() {
		double width = xmax - xmin;
		double height = ymax - ymin;
		double aspect = width/height;
		double windowAspect = (double)imageWidth/(double)imageHeight;
		if (aspect < windowAspect) {
			double newWidth = width*windowAspect/aspect;
			double center = (xmax + xmin)/2;
			xmax = center + newWidth/2;
			xmin = center - newWidth/2;
		}
		else if (aspect > windowAspect) {
			double newHeight = height*aspect/windowAspect;
			double center = (ymax+ymin)/2;
			ymax = center + newHeight/2;
			ymin = center - newHeight/2;
		}
	}
	

	public BufferedImage getImage() {
		return OSI;
	}
	public double getXmax() {
		return xmax;
	}
	public double getXmin() {
		return xmin;
	}
	public double getYmax() {
		return ymax;
	}
	public double getYmin() {
		return ymin;
	}
	
	private void makeOrbitData() {
		if (orbitType == MULTI_ORBIT) {
			double x1 = xmin + (double)startX/imageWidth*(xmax-xmin);
			double y1 = ymax - (double)startY/imageHeight*(ymax-ymin);
			double x2 = xmin + (double)endX/imageWidth*(xmax-xmin);
			double y2 = ymax - (double)endY/imageHeight*(ymax-ymin);
			double dx = (x2-x1)/(pointsOnOrbit-1);
			double dy = (y2-y1)/(pointsOnOrbit-1);
			double[] x = new double[pointsOnOrbit];
			double[] y = new double[pointsOnOrbit];
			for (int i = 0; i < pointsOnOrbit; i++) {
				x[i] = x1 + i*dx;
				y[i] = y1 + i*dy;
			}
			double[] zx = (double[])x.clone();
			double[] zy = (double[])y.clone();
			for (int line = 0; line < orbitLines; line++) {
				Vector lineData = new Vector();
				orbitData.addElement(lineData);
				for (int i = 0; i < pointsOnOrbit; i++) {
					if (Math.abs(zx[i]) < 100 && Math.abs(zy[i]) < 100)
						lineData.addElement(makePoint(zx[i],zy[i]));
					else
						lineData.addElement(new Point(0xFFFFFF,0));
				}
				if (line < orbitLines - 1) {
					for (int i = 0; i < pointsOnOrbit; i++) {
						double new_zx = zx[i]*zx[i] - zy[i]*zy[i] + x[i];
						zy[i] = 2*zx[i]*zy[i] + y[i];
						zx[i] = new_zx;
					}
				}
			}
		}
		else {
			orbitData.addElement(makePoint(X,Y));
			double zx = X;
			double zy = Y;
			for (int i = 1; i < pointsOnOrbit; i++) {
				double new_zx = zx*zx - zy*zy + X;
				zy = 2*zx*zy + Y;
				zx = new_zx;
				orbitData.addElement(makePoint(zx,zy));
				if (Math.abs(zx) > 5 || Math.abs(zy) > 5)
					break;
			}
		}
	}
	private Point makePoint(double x, double y) {
		int xint, yint;
		xint = (int)((x-xmin)/(xmax-xmin)*imageWidth+0.5);
		yint = (int)((ymax-y)/(ymax-ymin)*imageHeight+0.5);
		return new Point(xint,yint);
	}
	
	private void drawOrbits(Graphics g) {
		if (orbitType == NO_ORBIT)
			return;
		if (dragging) {
			g.setColor(Color.red);
			g.drawLine(startX,startY,endX,endY);
			return;
		}
		if (orbitData.size() == 0)
			makeOrbitData();
		if (orbitType == POINT_ORBIT) {
			for (int i = 0; i < orbitData.size(); i++) {
				Point point = (Point)orbitData.elementAt(i);
				int c = (int)(250.0*i/orbitData.size());
				g.setColor(orbitPalette[c]);
				g.fillOval(point.x-2,point.y-2,4,4);
			}
		}
		else if (orbitType == LINE_ORBIT) {
			Point point1 = (Point)orbitData.elementAt(0);
			for (int i = 1; i < orbitData.size(); i++) {
				Point point2 = (Point)orbitData.elementAt(i);
				int c = (int)(250.0*i/orbitData.size());
				g.setColor(orbitPalette[c]);
				g.drawLine(point1.x,point1.y,point2.x,point2.y);
				point1=point2;
			}
		}
		else if (orbitType == MULTI_ORBIT){
			for (int line = 0; line < orbitData.size(); line++) {
				Vector lineData = (Vector)orbitData.elementAt(line);
				int c = (int)(250.0*line/orbitData.size());
				g.setColor(orbitPalette[c]);
				Point point1 = (Point)lineData.elementAt(0);
				for (int i = 1; i < lineData.size(); i++) {
					Point point2 = (Point)lineData.elementAt(i);
					if (point1.x != 0xFFFFFF && point2.x != 0xFFFFFF)
						g.drawLine(point1.x,point1.y,point2.x,point2.y);
					point1 = point2;
				}
			}
		}
	}
	
	public void paintComponent(Graphics g) {
		super.paintComponent(g);
		if (OSI != null && (imageHeight != getHeight() || imageWidth != getWidth())) {
			comp.cancel();
			imageHeight = getHeight();
			imageWidth = getWidth();
			orbitType = NO_ORBIT;
			orbitData.setSize(0);
			makeImage();
			xmin = -2.333;
			xmax = 1.2;
			ymin = -1.25;
			ymax = 1.25;
			checkAspect();
			if (OSI != null)
				comp.install(this,500);
		}
		else if (OSI == null) {
			makeImage();
			if (OSI != null)
				comp.install(this,500);
		}
		if (OSI == null) {
			g.setColor(Color.gray);
			g.fillRect(0,0,getWidth(),getHeight());
			g.setColor(new Color(255,200,200));
			g.drawString("OUT OF MEMORY.  Try a smaller image??", 20,20);
		}
		else {
			g.setColor(Color.gray);
			if (imageWidth  < getWidth())
				g.fillRect(imageWidth,0,getWidth()-imageWidth,getHeight());
			if (imageHeight < getHeight())
				g.fillRect(0,imageHeight,imageWidth,getHeight()-imageHeight);
			g.drawImage(OSI,0,0,this);
			drawOrbits(g);
		}
	}
	
	private void makeImage() {
		OSI = null;
		try {
			//OSI = new BufferedImage(imageWidth,imageHeight,BufferedImage.TYPE_INT_RGB);
			IndexColorModel cm = new IndexColorModel(8,253,palette,0,false,
					-1,DataBuffer.TYPE_BYTE);
			OSI = new BufferedImage(imageWidth,imageHeight,BufferedImage.TYPE_BYTE_INDEXED,cm);
		}
		catch (OutOfMemoryError e) {
			OSI = null;
			return;
		}  
        Graphics g = OSI.getGraphics();
        g.setColor(Color.lightGray);
        g.fillRect(0,0,imageWidth,imageHeight);
        g.dispose();
	}
	
	synchronized public void doDraw() {
		forceStopDraw();
		if (OSI == null)
			makeImage();
		else {
			Graphics g = OSI.getGraphics();
			g.setColor(getBackground());
			g.fillRect(0,0,imageWidth,imageHeight);
			g.dispose();
		}
		repaint();
		if (OSI != null)
			comp.install(this);
	}
	
	synchronized public void stopDrawing() {
		comp.cancel();
	}
		
	public int getMaxIterations() {
		return maxIterations;
	}
	
	public void setMaxIterations(int iter) {
		maxIterations = iter;
	}
		
	public void newData(int x, int y, int width, int height) {
		repaint(x,y,width,height);
	}
	
	public boolean isRunning() {
		return comp.isRunning();
	}
	
	public void startedRunning() {
	}
	
	public void doneRunning() {
		// to be called from MandComp only
	}
	
	
	
}